import argparse
import cv2
import configparser
import logging
import os
import torch

from collections import defaultdict
from moviepy.editor import concatenate_videoclips, ImageClip
from pathlib import Path
from tqdm import tqdm


# import some common detectron2 utilities
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor
from detectron2.utils.visualizer import ColorMode, Visualizer

# Get ALDI imports
from aldi.config import add_aldi_config

# Define a dictionary mapping for logging levels
LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}

def is_float(value):
    """Check if the input value can be converted to a float."""
    try:
        float(value)
        return True
    except ValueError:
        return False

def is_int(value):
    """Check if the input value can be converted to an int."""
    try:
        int(value)
        return True
    except ValueError:
        return False

def setup_logging(level_name):
    '''Configure logging based on the level name provided by the user.'''
    
    # Set Log Level & Default to INFO if level is not recognized
    level = LOG_LEVELS.get(level_name.upper(), logging.INFO)  
    logging.basicConfig(level=level, 
                        format='%(asctime)s - %(levelname)s - %(message)s')
    logging.info(f"Logging level set to {level_name.upper()}")
    return None

class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

def get_config(config_file):
    '''Read configuration file and 
       return dataset, data_root_dir, output_root_dir, 
              model_cfg, model_weights, fps'''

    # Load configuration
    config = configparser.ConfigParser()
    config.read(config_file)

    # Check if required input & output path info is present
    try:
        dataset = config['data_path']['dataset']
        data_root_dir = config['data_path']['root_data_dir']
        output_root_dir = config['data_path']['root_movie_dir']
    except KeyError as e:
        logging.error(f"Missing information on input and output paths: {e}")
        sys.exit(1)

    # Check if required model configuration is present
    try:
        model_cfg = config['model_params']['cfg']
        model_weights = config['model_params']['weights']
    except KeyError as e:
        logging.error(f"Missing required configuration for model parameters: {e}")
        sys.exit(1)

    if is_float(config['model_params']['conf_threshold']):
        model_conf_threshold = float(config['model_params']['conf_threshold'])
    else:
        raise ValueError(f"conf_threshold must be a float. Got: {config['model']['conf_threshold']}")
    
    # Check if required frame rate info is present
    try:
        fps_str = config['movie_params']['fps'].strip()
        fps = int(fps_str)
    except KeyError as e:
        logging.info(f"Missing information on frame rate. Using default value of 10 fps")
        fps = 10
    except ValueError as e:
        logging.info(f"Non integer value for fps found - {fps_str}. Using default value of 10 fps")
        fps = 10

    return dataset, data_root_dir, output_root_dir, model_cfg, model_weights, fps
    

def get_predictor(expt_cfg, model_weights):
    '''Get Inference predictor for model with given config and weights'''
    cfg = get_cfg()
    cfg.set_new_allowed(True)
    add_aldi_config(cfg)
    cfg.merge_from_file(expt_cfg)
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5  # set threshold for this model
    cfg.MODEL.WEIGHTS = model_weights
    predictor = DefaultPredictor(cfg)
    return predictor


def make_frame_dict(src_dir_str):
    '''Return dict mapping each movie to its ordered frames'''
    src_dir = Path(src_dir_str)

    frame_dict = defaultdict(list)
    frame_list = sorted([x.name for x in src_dir.iterdir() 
                         if 'ipynb' not in x.name ])
    
    # Group images by prefix label
    for curr_img in frame_list:
        prefix = "_".join(curr_img.split("_")[0:9])
        frame_dict[prefix].append(curr_img)
    
    # Sort images for each prefix number by what I believe to be frame number
    for curr_movie in sorted(frame_dict.keys()):
        try:
            frame_dict[curr_movie] = sorted(frame_dict[curr_movie], 
                                            key=lambda x:int(x.split("_")[-1][:-4]))
        except ValueError as e:
            logging.warning(f"Failed to parse frame numbers in movie '{curr_movie}': {e}") 

    return frame_dict

def recover_pred_frame_dict(root_dir_str):
    pred_frame_dict = defaultdict(list)
    root_dir = Path(root_dir_str)

    # Get list of movies
    movie_list = sorted([x for x in root_dir.iterdir() 
                         if 'ipynb' not in x.name ])
    
    for curr_movie in movie_list:
        try:
            curr_frames = [x.name for x in curr_movie.iterdir()] 
        except:
            logging.error(f"Can't load image frames for {curr_movie}")
            continue
        
        curr_name = curr_movie.name
        temp = sorted(curr_frames, key=lambda x:int(x.split("_")[-1][:-4]))
        pred_frame_dict[curr_name] = [str(curr_movie / Path(x)) for x in temp]
        
    return pred_frame_dict

def rel_annotate_frame(image, text, relative_position=(.05,0.95), relative_font_size=0.001):
    """ Add text annotations to image using positions relative to image dimensions"""
    height, width = image.shape[:2]  # Get the image dimensions
    
    # Calculate the absolute font size and position relative to the image dimensions
    font_scale = relative_font_size * min(width, height)  # Scale font size based on image size
    x_pos = int(relative_position[0] * width)   # X position as a fraction of image width
    y_pos = int(relative_position[1] * height)  # Y position as a fraction of image height

    # Define other parameters for cv2.putText
    font = cv2.FONT_HERSHEY_SIMPLEX
    color = (255, 200, 200)  # Text color (change as needed)
    thickness = 1
    
    # Add the text to the image
    cv2.putText(image, text, (x_pos, y_pos), font, font_scale, 
                color, thickness, cv2.LINE_AA)

    return image

def make_pred_frame_dict(frame_dict, predictor, src_dir, out_dir, dataset):
    '''Return dict mapping each movie to its ordered prediction frames'''
    pred_frame_dict = defaultdict(list)
    num_movies = len(frame_dict)
    bad_file_list = []

    # Set metadata for Visualizer
    thing_colors = [(175, 175, 255)]
    metadata = AttributeDict({"thing_classes": ["Fish"],  
                              "thing_colors": thing_colors })
    
    # Loop through each movie
    for ctr, curr_movie_name in enumerate(frame_dict):
        # Make output directory for current movie
        curr_out_dir = out_dir / Path("Pred_Frames") / Path(dataset) /Path(curr_movie_name)
        curr_out_dir.mkdir(parents=True, exist_ok=True)

        logging.info(f"{curr_out_dir} {ctr+1} of {num_movies}")

        # Loop through each frame in movie
        for curr_img_name in tqdm(frame_dict[curr_movie_name]):

            # Get predictions for current image
            curr_img_path = os.path.join(src_dir, curr_img_name)
            im = cv2.imread(curr_img_path)
            try:
                outputs = predictor(im)
            except:
                logging.error(f"Error with {curr_img_name}")
                bad_file_list.append(curr_img_name)
                continue

            # Draw image
            v = Visualizer(im[:, :, ::-1], metadata=metadata,
                           scale=1.2, instance_mode=ColorMode.SEGMENTATION)
            out = v.draw_instance_predictions(outputs["instances"].to("cpu"))

            # Get image dimensions
            out_img = out.get_image()[:, :, ::-1]
            _, width, _ = out_img.shape

            # Calculate center coordinates for top-center position
            text_center_x = int(width / 2)
            text_center_y = 50  # Adjust y-coordinate as needed (distance from top edge)

            # Text parameters
            count = torch.gt(outputs['instances'].scores, 0).sum().item()
            text = "Fish Count: " + str(count)
            fontFace = cv2.FONT_HERSHEY_SIMPLEX
            fontScale = 0.65
            color = (255, 175, 175)  
            thickness = 1

            # Get text size using getTextSize
            (text_width, text_height), _ = cv2.getTextSize(text, fontFace, fontScale, thickness)

            # Adjust center_y based on text height for better alignment
            text_center_y = text_center_y + text_height // 2  # Add half of text height

            # Bottom-left corner coordinates relative to center
            text_org = (text_center_x - text_width // 2, text_center_y)  

            # Write text on image
            out_img=out_img.copy()
            cv2.putText(out_img, text, text_org, 
                        fontFace, fontScale, 
                        color, thickness, cv2.LINE_AA)
            
            # Write frame number on image
            curr_frame_num = int(curr_img_name.split("_")[-1][:-4])
            out_img = rel_annotate_frame(out_img, "Frame", (.865,0.895) )
            out_img = rel_annotate_frame(out_img, "Number", (.85,0.91) )
            out_img = rel_annotate_frame(out_img, str(curr_frame_num), (.90,0.925) )
            
            # Save image            
            outfile = curr_out_dir / Path("Pred_" + curr_img_name)
            cv2.imwrite(outfile, out_img)
            pred_frame_dict[curr_movie_name].append(str(outfile))
            
    if len(bad_file_list) > 0:
        logging.warning(f"Bad Files: {bad_file_list}" )
    return pred_frame_dict


def make_movies(pred_frame_dict, out_dir, dataset, fps=10):
    '''Create movies from predicted frames and save to out_dir'''

    # Make output directory
    out_dir_root = out_dir / Path(dataset)
    out_dir_root.mkdir(parents=True, exist_ok=True)
    num_movies = len(pred_frame_dict)


    #for ctr,curr_movie in enumerate(movie_list):
    for ctr,curr_movie in enumerate(pred_frame_dict):
        logging.info(f"{curr_movie}  {ctr+1} of {num_movies}")
        curr_movie_name = curr_movie.split("/")[-1]
        outfile_path = out_dir_root / Path(curr_movie_name + ".mp4")
        if outfile_path.exists():
            logging.warning(f" {outfile_path.name}  already exists and will not be replaced")
            continue
    
        # List image filenames
        images = pred_frame_dict[curr_movie]
        if len(images) == 0:
            logging.warning(f"{curr_movie}  has no images")
            break
    
        # Define clip duration (assuming all images have same duration)
        clip_duration = 1/fps  # Adjust clip duration per image (in seconds)
        
        # Create video clips from images
        clips = [ImageClip(image).set_duration(clip_duration) 
                 for image in images ]
    
        # Concatenate clips
        concat_clip = concatenate_videoclips(clips, method="compose")
        concat_clip.fps = fps

        logging.info(f"Saving {str(outfile_path)}")
        concat_clip.write_videofile(str(outfile_path),logger=None,codec = 'libx264')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config_file', default = "./movie_makers/annotated_movie.cfg", 
                        type=str, help="Path to the configuration file")
    parser.add_argument('-t', '--test', action='store_true')
    parser.add_argument('-r', '--recover_pred_frames', action='store_true')
    parser.add_argument('-l', '--log-level', default='INFO', choices=LOG_LEVELS.keys(),
                        help="Set the logging level (default: INFO)")
    args = parser.parse_args()

    # Set up logging
    setup_logging(args.log_level)
    
    # Check if configuration file exists
    config_file = args.config_file
    if Path(config_file).is_file():
        logging.info(f"Using configuration file: {config_file}")
    else:
        logging.error(f"Configuration file not found: {config_file}")
        sys.exit(1)

    # Load configuration
    config = configparser.ConfigParser()
    config.read(config_file)
     
    dataset, data_root_dir, output_root_dir, \
        model_cfg, model_weights, fps = get_config(config_file)

    if not args.recover_pred_frames:
        # Get predictor
        predictor = get_predictor(model_cfg, model_weights)

        # Make output directory
        out_dir = Path(output_root_dir) 
        out_dir.mkdir(parents=True, exist_ok=True)

        # Get frame dict
        src_dir = os.path.join(data_root_dir, dataset)
        frame_dict = make_frame_dict(src_dir)

        # Limit number of movies for testing
        if args.test:
            frame_dict = {k: frame_dict[k] for k in list(frame_dict)[0:5]}

        # Get predicted frame dict
        pred_frame_dict = make_pred_frame_dict(frame_dict, predictor, src_dir, out_dir, dataset)
    else:
        # Recover predicted frame dict
        logging.info("Recovering predicted frames and resuming movie making...")
        recover_dir = os.path.join(output_root_dir, 'Pred_Frames', dataset)
        pred_frame_dict = recover_pred_frame_dict(recover_dir)

    make_movies(pred_frame_dict, output_root_dir, dataset, fps)