import argparse
import configparser
import logging
import sys

from collections import defaultdict
from moviepy.editor import concatenate_videoclips, ImageClip
from pathlib import Path

# Define a dictionary mapping for logging levels
LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}

def setup_logging(level_name):
    '''Configure logging based on the level name provided by the user.'''
    
    level = LOG_LEVELS.get(level_name.upper(), logging.INFO)  # Default to INFO if level is not recognized
    logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(message)s')
    logging.info(f"Logging level set to {level_name.upper()}")
    return None

def get_config(config_file):
    '''Read & Validate configuration file. If valid,
       return dataset, data_root_dir, output_root_dir, and fps.
       Note: If fps is not present, default value of 10 is used'''
    # Load configuration
    config = configparser.ConfigParser()
    config.read(config_file)

    # Check if required input & output path info is present
    try:
        dataset = config['data_path']['dataset']
        data_root_dir = config['data_path']['root_data_dir']
        output_root_dir = config['data_path']['root_movie_dir']
    except KeyError as e:
        logging.critical(f"Missing information on input and output paths: {e}")
        sys.exit(1)

    # Check if required frame rate info is present
    try:
        fps = int(config['movie_params']['fps'])
    except KeyError as e:
        logging.warning(f"Missing information on frame rate. Using default value of 10 fps")
        fps = 10

    return dataset, data_root_dir, output_root_dir, fps

def make_frame_dict(src_dir_str):
    '''Return dict mapping each movie to its ordered frames. Assumes 
       frames are named as <prefix>_<frame_number>.<ext>'''
    src_dir = Path(src_dir_str)

    frame_dict = defaultdict(list)
    frame_list = sorted([x.name for x in src_dir.iterdir() 
                         if 'ipynb' not in x.name ])
    
    # Group images by prefix label
    for curr_img in frame_list:
        prefix = "_".join(curr_img.split("_")[:-1])
        frame_dict[prefix].append(curr_img)
    
    # Sort images for each prefix number by what I believe to be frame number
    for curr_movie in sorted(frame_dict.keys()):
        try:
            frame_dict[curr_movie] = sorted(frame_dict[curr_movie], 
                                            key=lambda x:int(x.split("_")[-1][:-4]))
        except ValueError as e:
            logging.warning(f"Failed to parse frame numbers in movie '{curr_movie}': {e}") 
    return frame_dict

def make_movies(frame_dict, src_dir, out_dir, dataset, fps):
    '''Create movies from frames and save to out_dir'''

    # Make output directory
    out_dir_root = out_dir / Path(dataset)
    out_dir_root.mkdir(parents=True, exist_ok=True)
    num_movies = len(frame_dict)
    logging.info(f"Creating {num_movies} movies")
    logging.info(f"Output directory: {out_dir_root}")

    for ctr,curr_movie in enumerate(frame_dict):
        logging.info(f"Processing {curr_movie} ({ctr + 1} of {num_movies})")
        curr_movie_name = curr_movie.split("/")[-1]
        outfile_path = out_dir_root / Path(curr_movie_name + ".mp4")
        if outfile_path.exists():
            logging.info(f" {outfile_path.name}  already exists")
            continue
    
        # List image filenames
        images = frame_dict[curr_movie]
        if len(images) == 0:
            logging.warning(f"{curr_movie_name} has no images")
            break
    
        # Define clip duration (time each image is shown for)
        clip_duration = 1/fps  # Adjust clip duration per image (in seconds)
        
        # Create video clips from images
        try:
            clips = [
                ImageClip(str(Path(src_dir) / image)).set_duration(clip_duration)
                for image in images
            ]
            concat_clip = concatenate_videoclips(clips, method="compose")
            concat_clip.fps = fps
            logging.info(f"Saving {outfile_path}")
            concat_clip.write_videofile(str(outfile_path), logger=None, codec='libx264')
        except Exception as e:
            logging.error(f"Error creating video for {curr_movie}: {e}")
            continue


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config_file', 
                        default = "./movie_makers/raw_movie.cfg", 
                        type=str, help="Path to the configuration file")
    parser.add_argument('-t', '--test', action='store_true')
    parser.add_argument('-l', '--log-level', default='INFO', choices=LOG_LEVELS.keys(),
                        help="Set the logging level (default: INFO)")
    
    args = parser.parse_args()

    # Set up logging
    setup_logging(args.log_level)
    
    # Check if configuration file exists
    config_file = args.config_file
    if Path(config_file).is_file():
        logging.info(f"Using configuration file: {config_file}")
    else:
        logging.error(f"Configuration file not found: {config_file}")
        sys.exit(1)

    # Load configuration
    dataset, data_root_dir, output_root_dir, fps = get_config(config_file)
    
    # Make output directory
    out_dir = Path(output_root_dir) / Path(dataset)
    out_dir.mkdir(parents=True, exist_ok=True)

    # Get frame dict
    src_dir = Path(data_root_dir) / Path(dataset)
    frame_dict = make_frame_dict(src_dir)

    # Limit number of movies for testing
    if args.test:
        frame_dict = {k: frame_dict[k] for k in list(frame_dict)[0:5]}
    
    make_movies(frame_dict, src_dir, output_root_dir, dataset, fps)
    logging.info("Done")
