import argparse
import configparser
import copy
import cv2
import importlib
import logging
import sys

from enum import Enum
from pathlib import Path

# Get ALDI imports
from aldi.config import add_aldi_config

# Get Detectron2 imports
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor

# Get Norfair imports
from norfair import Detection, Tracker, Video
from norfair import draw_points, print_objects_as_table
from norfair.distances import _SCALAR_DISTANCE_FUNCTIONS, AVAILABLE_VECTORIZED_DISTANCES

# Add directory containing 'norfair_custom_utils.py' to sys.path for import
norfair_utils_dir = Path(__file__).resolve().parent / 'norfair_custom_utils'
sys.path.append(str(norfair_utils_dir))

# My Norfair custom module
from norfair_custom_utils import fileprint_objects_as_table

# Created list of distance functions pre-defined in norfair
NORFAIR_DISTANCE_FUNCTIONS = copy.deepcopy(AVAILABLE_VECTORIZED_DISTANCES)
NORFAIR_DISTANCE_FUNCTIONS.extend(_SCALAR_DISTANCE_FUNCTIONS)

# Declare module where Norfair tracking 
# utility functions are defined
TRACKER_UTILITY_MODULE = "track_movie_utils"

class TrackerConfigValue(Enum):
    DEFAULT = 'default'
    NONE = 'none'

# Define a dictionary mapping for logging levels
LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}

def is_float(value):
    """Check if the input value can be converted to a float."""
    try:
        float(value)
        return True
    except ValueError:
        return False

def is_int(value):
    """Check if the input value can be converted to an int."""
    try:
        int(value)
        return True
    except ValueError:
        return False

def setup_logging(level_name):
    '''Configure logging based on the level name provided by the user.'''
    
    level = LOG_LEVELS.get(level_name.upper(), logging.INFO)  # Default to INFO if level is not recognized
    logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(message)s')
    logging.info(f"Logging level set to {level_name.upper()}")
    return None

def get_config(config_file):
    '''Read & Validate configuration file. If valid,
       return model_cfg, model_weights, input_dir, output_dir'''
    
    # Load configuration
    config = configparser.ConfigParser()
    config.read(config_file)

    # Check if required model configuration is present
    try:
        model_cfg = config['model']['cfg']
        model_weights = config['model']['weights']
        if is_float(config['model']['conf_threshold']):
            model_conf_threshold = float(config['model']['conf_threshold'])
        else:
            raise ValueError(f"conf_threshold must be a float. Got: {config['model']['conf_threshold']}")
    except KeyError as e:
        logging.error(f"Missing required configuration for model parameters: {e}")
        sys.exit(1)

    # Check if required path configuration is present
    try:
        input_dir = Path(config['data_path']['input_dir'])
        output_dir = Path(config['data_path']['output_dir'])
    except KeyError as e:
        logging.error(f"Missing required configuration key: {e}")
        sys.exit(1)

    # Verify input directory exists
    if not input_dir.is_dir():
        logging.error(f"Movie directory not found: {input_dir}")
        sys.exit(1)

    # Check if required frame rate info is present
    try:
        fps = int(config['movie_params']['fps'])
    except KeyError as e:
        logging.warning(f"Missing information on frame rate. Using default value of 10 fps")
        fps = 10

    # Get dict of Tracker arguments from config file
    tracker_args = dict(config['tracker'])

    try:
        debug_movie_list = [x.strip() for x in config['debug']['debug_movies'].split(',')]
    except KeyError:
        debug_movie_list = []
    logging.info(f"Debug movies: {debug_movie_list}")
    
    return (model_cfg, model_weights, model_conf_threshold, 
            input_dir, output_dir, fps, tracker_args, debug_movie_list)

def get_detector(expt_cfg, model_weights, model_conf_threshold):
    '''Get Inference predictor for model with given config and weights'''
    cfg = get_cfg()
    cfg.set_new_allowed(True)
    add_aldi_config(cfg)
    cfg.merge_from_file(expt_cfg)
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = model_conf_threshold  # set threshold for this model
    cfg.MODEL.WEIGHTS = model_weights
    detector = DefaultPredictor(cfg)
    logging.info(f"Loaded model with config: {expt_cfg} and weights: {model_weights}")
    logging.info(f"Confidence threshold set to: {model_conf_threshold}")
    return detector

# Helper function to import a function from the default module
def import_function(function_name):
    try:
        module = importlib.import_module(TRACKER_UTILITY_MODULE)  # Import the custom module
    except ModuleNotFoundError  as e:
        logging.error(f"Error importing module: {e}")
        
    try:
        func = getattr(module, function_name)  # Get the function from the module   
    except AttributeError as e: 
        logging.error(f"Error importing function: {e}") 
        func = None
    return func
    #module = importlib.import_module(TRACKER_UTILITY_MODULE)  # Import the default module
    #return getattr(module, function_name)  # Get the function from the module

def draw_text(frame, text, text_center_x, text_center_y, 
              font_face=cv2.FONT_HERSHEY_SIMPLEX, 
              font_scale=0.65, text_color=(255,200,200), 
              text_thickness=1):
    '''Draws centered text on the frame at the specified position
       frame is modified in-place. No need to return anything'''

    # Get text size using getTextSize
    (text_width, text_height), _ = cv2.getTextSize(text, font_face, 
                                                   font_scale, 
                                                   text_thickness)

    # Adjust center_y based on text height for better alignment
    text_center_y = text_center_y + text_height // 2  # Add half of text height

    # Bottom-left corner coordinates relative to center
    text_org = (text_center_x - text_width // 2, text_center_y)  

    # Write text on image
    cv2.putText(frame, text, text_org, 
                font_face, font_scale, 
                text_color, text_thickness, 
                cv2.LINE_AA)
    return None

def rel_annotate_frame(image, text, relative_position=(.05,0.95), relative_font_size=0.001):
    """ Add text annotations to image using positions relative to image dimensions"""
    height, width = image.shape[:2]  # Get the image dimensions
    
    # Calculate the absolute font size and position relative to the image dimensions
    font_scale = relative_font_size * min(width, height)  # Scale font size based on image size
    x_pos = int(relative_position[0] * width)   # X position as a fraction of image width
    y_pos = int(relative_position[1] * height)  # Y position as a fraction of image height

    # Define other parameters for cv2.putText
    font = cv2.FONT_HERSHEY_SIMPLEX
    color = (255, 200, 200)  # Text color (change as needed)
    thickness = 1
    
    # Add the text to the image
    cv2.putText(image, text, (x_pos, y_pos), font, font_scale, 
                color, thickness, cv2.LINE_AA)

    return image

def parse_tracker_args(tracker_args):
    '''Parse tracker arguments from config file'''
    del_list = []
    for key, value in tracker_args.items():
        if key == 'distance_function':
            # Note: This function is imported from the default module
            # it is a required input for the tracker
            if value in NORFAIR_DISTANCE_FUNCTIONS:
                logging.info(f"Using Norfair-defined distance function: {value}")
                tracker_args[key] = value
            else:
                tracker_args[key] = import_function(value)
                logging.info(f"Using custom distance function: {value}")
            continue
        elif value.lower() == TrackerConfigValue.DEFAULT.value:
            del_list.append(key)
            continue            
        elif value.lower() == TrackerConfigValue.NONE.value:
            tracker_args[key] = None
            continue
        elif is_float(value):
            tracker_args[key] = float(value)
            continue    
        elif is_int(value):        
            tracker_args[key] = int(value)
            continue
        elif key == 'reid_distance_function':
            # Note: This function is imported from the default module
            # it is not required input for the tracker, so None value
            # is checked for first.
            tracker_args[key] = import_function(value)
            continue
        else:
            logging.error(f"Invalid value for {key}: {value}")
            sys.exit(1)
            
    for key in del_list:
        del tracker_args[key]   
         
    return tracker_args

def make_tracking_video(model_cfg, model_weights, model_conf_threshold, 
                        input_dir, output_dir, fps, tracker_args, debug_movie_list):
    '''Add information to existing video using Detectron2 & Norfair tracker'''

    # Get object detector
    detector = get_detector(model_cfg, model_weights, model_conf_threshold)

    # Create output directory, including any parent directories if necessary
    output_dir.mkdir(parents=True, exist_ok=True)

    # Get list of movies in input directory
    movies = [f for f in input_dir.iterdir() 
              if f.is_file() and f.suffix == '.mp4']

    # Check if there are any .mp4 files to process
    if not movies:
        logging.warning(f"No .mp4 files found in {input_dir}")
        return

    # Process each movie
    for curr_movie in movies:
        track_name = curr_movie.stem + "_tracked.mp4"
        tracked_movie = output_dir / track_name

        # Initialize video writer and tracker
        video = Video(input_path=str(curr_movie),
                      output_path=str(tracked_movie),
                      output_fps=fps)

        tracker = Tracker(**tracker_args)
        
        unique_fish_ids = set()
        for ctr,frame in enumerate(video):
            
            # Get detections in frame
            detections = detector(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            detections = [Detection(p) 
                        for p in detections['instances'].pred_boxes.get_centers().cpu().numpy()]

            # Update tracker with detections
            tracked_objects = tracker.update(detections=detections)
            for curr in tracked_objects:
                unique_fish_ids.add(curr.initializing_id)
            
            # Draw tracked objects on frame
            draw_points(frame, tracked_objects)
            
            # Get image dimensions
            _, width, _ = frame.shape

            # Calculate center coordinates for top-center position
            text_center_x = int(width / 2)

            # Draw 1st line of text on frame
            text_center_y = 25  # Adjust y-coordinate as needed (distance from top edge)
            text = "Total Number of Fish: "
            draw_text(frame, text, text_center_x, text_center_y)
            
            # Draw 2nd line of text on frame
            text_center_y = 50  # Adjust y-coordinate as needed (distance from top edge)
            text = str(len(unique_fish_ids))
            draw_text(frame, text, text_center_x, text_center_y)

            # Write frame number on image
            curr_frame_num = ctr
            frame = rel_annotate_frame(frame, "Frame", (.865,0.895) )
            frame = rel_annotate_frame(frame, "Number", (.85,0.91) )
            frame = rel_annotate_frame(frame, str(curr_frame_num), (.90,0.925) )
            
            # Write frame to video
            video.write(frame)       

            # Print tracked objects as table to file 
            if curr_movie.stem in debug_movie_list:
                fileprint_objects_as_table(tracked_objects, 
                                           output_file=tracked_movie.with_suffix('.txt'), 
                                           frame_number=ctr)
        

# Main function
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Track fish in a video")
    parser.add_argument('-c', '--config_file', default = "./movie_makers/track_movie.cfg", 
                        type=str, help="Path to the configuration file")
    parser.add_argument('-l', '--log-level', default='INFO', choices=LOG_LEVELS.keys(),
                        help="Set the logging level (default: INFO)")
    args = parser.parse_args()

    # Set up logging
    setup_logging(args.log_level)
  
    # Check if configuration file exists
    config_file = args.config_file
    if Path(config_file).is_file():
        logging.info(f"Using configuration file: {config_file}")
    else:
        logging.error(f"Configuration file not found: {config_file}")
        sys.exit(1)
        
    # Load configuration
    (model_cfg, model_weights, model_conf_threshold, 
     input_dir, output_dir, fps, raw_tracker_args, debug_movie_list) = get_config(config_file)
    tracker_args = parse_tracker_args(raw_tracker_args)
    
    make_tracking_video(model_cfg, model_weights, model_conf_threshold,
                        input_dir, output_dir, fps, tracker_args, debug_movie_list)
    