import cv2
import matplotlib.pyplot as plt
import numpy as np
import sys
from time import time

from collections import defaultdict
from IPython.display import display
from pathlib import Path
from PIL import Image, ImageDraw, ImageEnhance
from scipy.ndimage import gaussian_filter, uniform_filter
from skimage import exposure, io, restoration
from skimage.util import view_as_blocks, view_as_windows

# import some common detectron2 utilities
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor

# Get ALDI imports
from aldi.config import add_aldi_config

def np_stats_str(arr):
    return f"{np.min(arr):.4f}, {np.mean(arr):.4f}, {np.max(arr):.4f}"

def get_predictor(expt_cfg, model_weights):
    '''Get Inference predictor for model with given config and weights'''
    cfg = get_cfg()
    cfg.set_new_allowed(True)
    add_aldi_config(cfg)
    cfg.merge_from_file(expt_cfg)
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5  # set threshold for this model
    cfg.MODEL.WEIGHTS = model_weights
    predictor = DefaultPredictor(cfg)
    return predictor

def get_gnd_truth_bboxes(image_path, predictor, conf_thresh=0):
    '''Need to rename this function. It gets predicted bounding boxes
       if confidence is high enough'''

    image = cv2.imread(image_path)
    outputs = predictor(image)
    #print("outputs:", outputs)

    # Bbox predictions & scores
    pred_boxes = outputs['instances'].pred_boxes.tensor.cpu().numpy()
    pred_scores = outputs['instances'].scores.cpu().numpy()

    # Only accept bboxes with conf score above threshold
    pred_boxes_out = []
    for curr_pb, curr_score in zip(pred_boxes, pred_scores):
        if curr_score > conf_thresh:
            pred_boxes_out.append(curr_pb)
    return pred_boxes_out

def get_gnd_truth_bboxes2(image, predictor, conf_thresh=0):
    '''Need to rename this function. It gets predicted bounding boxes
       if confidence is high enough'''

    outputs = predictor(image)
    #print("outputs:", outputs)

    # Bbox predictions & scores
    pred_boxes = outputs['instances'].pred_boxes.tensor.cpu().numpy()
    pred_scores = outputs['instances'].scores.cpu().numpy()

    # Only accept bboxes with conf score above threshold
    pred_boxes_out = []
    for curr_pb, curr_score in zip(pred_boxes, pred_scores):
        if curr_score > conf_thresh:
            pred_boxes_out.append(curr_pb)
    return pred_boxes_out

def get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col):

    # Determine start & end rows for current grid cell
    if curr_row + grid_size <= num_rows:
        start_row = curr_row
        end_row = start_row + grid_size
    else:
        start_row = curr_row #num_rows - grid_size
        end_row = num_rows

    # Determine start and end cols for current grid cell
    if curr_col + grid_size <= num_cols:
        start_col = curr_col
        end_col = start_col + grid_size
    else:
        start_col = curr_col #num_cols - grid_size
        end_col = num_cols

    return start_col, end_col, start_row, end_row

def orig_image(image):
    # Dummy function to test preprocess
    return image

def remove_big_fish(image_np, pred_boxes, grid_size=1):
    # Remove large fish from image, so that only fish left are fry(?)

    big_fish_free = np.copy(image_np)
    # Separate the color channels (Raw, Background, Frame Diff)
    raw_channel = big_fish_free[:, :, 0]
    bckgnd_channel = big_fish_free[:, :, 1]
    dframe_channel = big_fish_free[:, :, 2]

    # Remove Detected Fish
    for curr_pred_box in pred_boxes:
        x1,y1,x2,y2 = curr_pred_box

        
        if grid_size > 0:
            # Get grid over which to average pixel values
            x1b = int(grid_size * np.floor(x1/grid_size))
            x2b = int(grid_size * np.ceil(x2/grid_size))
            y1b = int(grid_size * np.floor(y1/grid_size))
            y2b = int(grid_size * np.ceil(y2/grid_size))

            # Get average pixel values over chosen grid
            raw_chan_mean = np.mean(raw_channel[y1b:y2b,x1b:x2b])
            bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b,x1b:x2b])
            dframe_chan_mean = np.mean(dframe_channel[y1b:y2b,x1b:x2b])
            #print(raw_chan_mean, bckgnd_chan_mean, dframe_chan_mean)
        else:
            # Get avg pixel values over bbox
            raw_chan_mean = np.mean(raw_channel[y1:y2,x1:x2])
            bckgnd_chan_mean = np.mean(bckgnd_channel[y1:y2,x1:x2])
            dframe_chan_mean = np.mean(dframe_channel[y1:y2,x1:x2])
            #print(raw_chan_mean, bckgnd_chan_mean, dframe_chan_mean)

        # Replace detected fish with average pixel values
        raw_channel [int(np.floor(y1)):int(np.ceil(y2)),
                     int(np.floor(x1)):int(np.ceil(x2))] = raw_chan_mean
        bckgnd_channel [int(np.floor(y1)):int(np.ceil(y2)),
                        int(np.floor(x1)):int(np.ceil(x2))] = bckgnd_chan_mean
        dframe_channel [int(np.floor(y1)):int(np.ceil(y2)),
                        int(np.floor(x1)):int(np.ceil(x2))] = dframe_chan_mean
        
    big_fish_free_np = np.stack((raw_channel, bckgnd_channel, dframe_channel), axis=-1)
    return big_fish_free_np

def bilateral_filter_1D(image, sigma_color, sigma_spatial):
    
    # Initialize the filtered image
    filtered_image = np.zeros_like(image)
    
    # Define the size of the filter
    size = int(2 * sigma_spatial + 1)
    
    # Apply the bilateral filter
    for i in range(image.shape[0]):
        for j in range(image.shape[1]):
            # Define the region around the pixel
            i_min = max(0, i - size // 2)
            i_max = min(image.shape[0], i + size // 2 + 1)
            j_min = max(0, j - size // 2)
            j_max = min(image.shape[1], j + size // 2 + 1)
            
            region = image[i_min:i_max, j_min:j_max]
            
            # Calculate spatial weight
            spatial_distance = np.sqrt((np.arange(i_min, i_max)[:, None] - i)**2 + (np.arange(j_min, j_max) - j)**2)
            spatial_weight = np.exp(-0.5 * spatial_distance**2 / (sigma_spatial**2))
            
            # Calculate color weight
            color_weight = np.exp(-0.5 * (region - image[i, j])**2 / (sigma_color**2))
            
            # Combine spatial and color weights
            weight = spatial_weight * color_weight
            
            # Normalize weights
            weight_sum = np.sum(weight)
            if weight_sum > 0:
                filtered_image[i, j] = np.sum(region * weight) / weight_sum
            else:
                filtered_image[i, j] = image[i, j]
    
    return filtered_image

def bilateral_filter_3D(image_np, sigma_color, sigma_spatial):

    # Separate the color channels (Raw, Background, Frame Diff)
    image_np = image_np.astype(np.float64)
    raw_channel = image_np[:, :, 0]
    bckgnd_channel = image_np[:, :, 1]
    dframe_channel = image_np[:, :, 2]
    
    # Apply Bilateral Filter
    bilateral_raw_np = bilateral_filter_1D(raw_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)
    bilateral_bckgnd_np = bilateral_filter_1D(bckgnd_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)
    bilateral_dframe_np = bilateral_filter_1D(dframe_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)

    bilateral_np = np.stack((bilateral_raw_np, bilateral_bckgnd_np, bilateral_dframe_np), axis=-1)
    bilateral_np = bilateral_np.astype(np.uint8)

    return bilateral_np

def bilateral_filter_3D(image_np, sigma_color, sigma_spatial):

    # Separate the color channels (Raw, Background, Frame Diff)
    image_np = image_np.astype(np.float64)
    raw_channel = image_np[:, :, 0]
    bckgnd_channel = image_np[:, :, 1]
    dframe_channel = image_np[:, :, 2]
    
    # Apply Bilateral Filter
    bilateral_raw_np = bilateral_filter_1D(raw_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)
    bilateral_bckgnd_np = bilateral_filter_1D(bckgnd_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)
    bilateral_dframe_np = bilateral_filter_1D(dframe_channel, sigma_color=sigma_color, sigma_spatial=sigma_spatial)

    bilateral_np = np.stack((bilateral_raw_np, bilateral_bckgnd_np, bilateral_dframe_np), axis=-1)
    bilateral_np = bilateral_np.astype(np.uint8)

    return bilateral_np

def guided_filter_channel(I, p, radius, eps):
    """
    Guided filter applied to a single channel.
    
    Parameters:
    - I: Guidance image (H, W), single channel.
    - p: Input image to be filtered (H, W), single channel.
    - radius: Radius of the filter window.
    - eps: Regularization term.
    
    Returns:
    - q: Filtered output image (H, W).
    """
    mean_I = uniform_filter(I, size=radius)
    mean_p = uniform_filter(p, size=radius)
    mean_Ip = uniform_filter(I * p, size=radius)
    cov_Ip = mean_Ip - mean_I * mean_p

    mean_II = uniform_filter(I * I, size=radius)
    var_I = mean_II - mean_I * mean_I

    a = cov_Ip / (var_I + eps)
    b = mean_p - a * mean_I

    mean_a = uniform_filter(a, size=radius)
    mean_b = uniform_filter(b, size=radius)

    q = mean_a * I + mean_b
    return q

def vectorized_guided_filter_rgb_independent(img, radius=5, eps=0.01):
    """
    Apply vectorized guided filtering independently to each channel of an RGB image.

    Parameters:
    - img: Input RGB image (H, W, 3)
    - radius: Radius of the filter
    - eps: Regularization term
    
    Returns:
    - Filtered image (H, W, 3)
    """
    img = img.astype(np.float32)  # Convert image to float32 for computation
    
    # Initialize output image
    filtered_img = np.zeros_like(img)

    # Apply guided filter on each channel independently
    for i in range(3):
        filtered_img[:, :, i] = guided_filter_channel(img[:, :, i], img[:, :, i], radius, eps)

    return np.clip(filtered_img, 0, 255).astype(np.uint8)


def vectorized_guided_filter_1D(img, radius=5, eps=0.01):
    """
    Apply vectorized guided filtering independently to each channel of an RGB image.

    Parameters:
    - img: Input RGB image (H, W, 3)
    - radius: Radius of the filter
    - eps: Regularization term
    
    Returns:
    - Filtered image (H, W, 3)
    """
    img = img.astype(np.float32)  # Convert image to float32 for computation
    
    # Initialize output image
    filtered_img = np.zeros_like(img)

    # Apply guided filter 
    filtered_img[:, :] = guided_filter_channel(img[:, :], img[:, :], radius, eps)

    return np.clip(filtered_img, 0, 255).astype(np.uint8)



def clahe_filter(image_np, clip_limit):

    # Separate the color channels (Raw, Background, Frame Diff)
    raw_channel = image_np[:, :, 0]
    bckgnd_channel = image_np[:, :, 1]
    dframe_channel = image_np[:, :, 2]

    # Apply CLAHE to each channel
    clahe_raw = exposure.equalize_adapthist(raw_channel, clip_limit=clip_limit)
    clahe_bckgnd = exposure.equalize_adapthist(bckgnd_channel, clip_limit=clip_limit)
    clahe_dframe = exposure.equalize_adapthist(dframe_channel, clip_limit=clip_limit)

    # Convert back to 3Plane Numpy
    clahe_np = np.stack((clahe_raw, clahe_bckgnd, clahe_dframe), axis=-1)
    clahe_np *= 255
    clahe_np = clahe_np.astype(np.uint8)

    return clahe_np


def image_enhance_1D(image_np, enhance_factor):

    # Convert numpy to PIL Images
    img_pil = Image.fromarray(image_np.astype(np.uint8))
    
    # Apply ImageEnhance to each CLAHE-enhanced channel
    enhancer_pil = ImageEnhance.Contrast(img_pil)
    
    enhanced_pil = enhancer_pil.enhance(enhance_factor)
    
    # Convert enhanced PIL Images back to NumPy arrays
    enhanced_pil_np = np.array(enhanced_pil)

    return enhanced_pil

def image_enhance_3D(image_np, enhance_factor):

    # Separate the color channels (Raw, Background, Frame Diff)
    raw_channel = image_np[:, :, 0]
    bckgnd_channel = image_np[:, :, 1]
    dframe_channel = image_np[:, :, 2]

    # Convert enhanced PIL Images back to NumPy arrays
    enhanced_raw_np = image_enhance_1D(raw_channel, enhance_factor)
    enhanced_bckgnd_np = image_enhance_1D(bckgnd_channel, enhance_factor)
    enhanced_dframe_np = image_enhance_1D(dframe_channel, enhance_factor)

    enhanced_np = np.stack((enhanced_raw_np, enhanced_bckgnd_np, enhanced_dframe_np), axis=-1)

    return enhanced_np


def image_enhance(image_np, enhance_factor):

    # Separate the color channels (Raw, Background, Frame Diff)
    raw_channel = image_np[:, :, 0]
    bckgnd_channel = image_np[:, :, 1]
    dframe_channel = image_np[:, :, 2]

    # Convert numpy to PIL Images
    raw_pil = Image.fromarray(raw_channel.astype(np.uint8))
    bckgnd_pil = Image.fromarray(bckgnd_channel.astype(np.uint8))
    dframe_pil = Image.fromarray(dframe_channel.astype(np.uint8))
    
    # Apply ImageEnhance to each CLAHE-enhanced channel
    enhancer_raw = ImageEnhance.Contrast(raw_pil)
    enhancer_bckgnd = ImageEnhance.Contrast(bckgnd_pil)
    enhancer_dframe = ImageEnhance.Contrast(dframe_pil)
    
    enhanced_raw = enhancer_raw.enhance(enhance_factor)
    enhanced_bckgnd = enhancer_bckgnd.enhance(enhance_factor)
    enhanced_dframe = enhancer_dframe.enhance(enhance_factor)
    
    # Convert enhanced PIL Images back to NumPy arrays
    enhanced_raw_np = np.array(enhanced_raw)
    enhanced_bckgnd_np = np.array(enhanced_bckgnd)
    enhanced_dframe_np = np.array(enhanced_dframe)

    enhanced_np = np.stack((enhanced_raw_np, enhanced_bckgnd_np, enhanced_dframe_np), axis=-1)

    return enhanced_np


def nth_largest_element_wrapper(n):
    """
    Create a function that returns the nth largest element from a NumPy array.
    
    Parameters:
    - n: The position of the largest value to find (e.g., 1 for largest, 2 for second largest).
    
    Returns:
    - A function that takes a NumPy array as input and returns the nth largest element.
    """
    def nth_largest(arr):
        """
        Find the nth largest value in the array using np.partition.
        
        Parameters:
        - arr: Input NumPy array (1D).
        
        Returns:
        - The nth largest value.
        """
        if not (1 <= n <= len(arr)):
            raise ValueError("n must be between 1 and the length of the array.")
        
        # Perform partitioning to find the nth largest element
        partitioned_arr = np.partition(arr.ravel(), -n)
        
        # Return the nth largest value
        return partitioned_arr[-n]
    
    return nth_largest


def np_filter_function(image_np, np_func, grid_size, thresholds, 
                       enhance_factor, radius, eps,
                       verbose=False):
    
    image_np = image_np.astype(np.float64)
    new_image_np = np.copy(image_np)

    if isinstance(thresholds, (list, tuple)):
        if len(thresholds) !=3:
            print(f"Expected threshold list to have len == 3, not {len(thresholds)}")
            print(f"thresholds = {thresholds}")
            return
    elif isinstance(thresholds, (int, float, complex, np.number)):
        thresholds = [thresholds]*3

    # Prepare lists for raw, bckgnd and dframe filter paramss values
    filter_param_lists = [[], [], []]

    # 1st scan to get filter params values
    if verbose:
        print("Printing orig filter params")
    num_rows, num_cols, _ = image_np.shape
    for curr_row in range(0,num_rows,grid_size):
        for curr_col in range(0,num_cols,grid_size):
    

            # Determine start & end rows for current grid cell
            if curr_row + grid_size <= num_rows:
                start_row = curr_row
                end_row = start_row + grid_size
            else:
                start_row = curr_row #num_rows - grid_size
                end_row = num_rows
        
            # Determine start and end cols for current grid cell
            if curr_col + grid_size <= num_cols:
                start_col = curr_col
                end_col = start_col + grid_size
            else:
                start_col = curr_col #num_cols - grid_size
                end_col = num_cols

    
            curr_cell = image_np[start_row:end_row, start_col:end_col,:]
    
            #Get filter params for current grid cell
            curr_cell_filter_params = [np_func(curr_cell[:,:,ctr]) for ctr in range(3)]
    
            #Save to lists
            for ctr in range(3):
                filter_param_lists[ctr].append(curr_cell_filter_params[ctr])
 
    # 2nd scan to filter
    for curr_row in range(0,num_rows,grid_size):
        for curr_col in range(0,num_cols,grid_size):    

            # Determine start & end rows for current grid cell
            if curr_row + grid_size <= num_rows:
                start_row = curr_row
                end_row = start_row + grid_size
            else:
                start_row = curr_row #num_rows - grid_size
                end_row = num_rows
        
            # Determine start and end cols for current grid cell
            if curr_col + grid_size <= num_cols:
                start_col = curr_col
                end_col = start_col + grid_size
            else:
                start_col = curr_col #num_cols - grid_size
                end_col = num_cols


            curr_cell = image_np[start_row:end_row, start_col:end_col,:]
    
            #Get filter params for current grid cell
            curr_cell_filter_params = [np_func(curr_cell[:,:,i]) for i in range(3)]
    
            for i, (curr_filter_param, curr_thresh) in enumerate(zip(curr_cell_filter_params, thresholds)):
                if curr_filter_param > curr_thresh:
                    # Increase contrast of entire cell if above threshold
                    new_image_np[start_row:end_row, start_col:end_col, i] = image_enhance_1D(curr_cell[:,:,i], 
                                                                                             enhance_factor=enhance_factor)
                else:
                    new_image_np[start_row:end_row, start_col:end_col, i] = np.mean(curr_cell[:,:,i])
                    # Smooth cell if beneath threshold
                    new_image_np[start_row:end_row, start_col:end_col, i] = \
                        vectorized_guided_filter_1D(curr_cell[:,:,i], radius=radius, eps=eps)

                    #new_image_np[start_row:end_row, start_col:end_col, i] = bilateral_filter_1D(curr_cell[:,:,i], 
                    #                                                                            sigma_color=sigma_color, 
                    #                                                                            sigma_spatial=sigma_spatial)

    new_image_np = new_image_np.astype(np.uint8)
    return new_image_np            


def remove_dark_noise(image_np, threshold_perc_list, frame_list):

    for curr_thresh_perc, curr_frame in zip(threshold_perc_list, frame_list):
        thresh = np.percentile(image_np[:,:,curr_frame], curr_thresh_perc)
        #print(f"{np_stats_str(image_np[750:800,0:50,1])}")
        #print(f" Frame {curr_frame} thresh = {thresh} orig_min = {np.min(image_np[:,:,curr_frame])}", end=" ")
        image_np[:,:,curr_frame][image_np[:,:,curr_frame] < thresh] = int(thresh)
        #print(f" new_min = {np.min(image_np[:,:,curr_frame])}")
        #print(f"{np_stats_str(image_np[750:800,0:50,1])}")
        #image_np[750:800,0:50,1] = int(127)

    return image_np

def get_gradient_magnitude_image(image):
    
    # Apply Sobel operators
    sobel_x = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
    sobel_y = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)
    
    # Compute and normalize the gradient magnitude
    gradient_magnitude = np.sqrt(sobel_x**2 + sobel_y**2)
    gradient_magnitude = cv2.normalize(gradient_magnitude, None, 0, 255, cv2.NORM_MINMAX)
    gradient_magnitude.astype(np.uint8)

    return gradient_magnitude


def edge_detected_image(image, canny_lower, canny_upper):
    
    # Perform edge detection on bckgnd frame & replace orig and delta frames with edge_img
    edges = cv2.Canny(image[:,:,1], canny_lower, canny_upper)    
    edge_img = np.stack((edges, edges, edges), axis=-1)

    return edge_img


def get_edgy_ims(image, canny_lower, canny_upper):

    # Focus more on analysis of edges in image
    gradient_magnitude = get_gradient_magnitude_image(image[:,:,1])
    edge_img = cv2.Canny(image[:,:,1], canny_lower, canny_upper)   
    edgy_ims = np.stack((image[:,:,1], gradient_magnitude, edge_img), axis=-1)
    edgy_ims = np.clip(edgy_ims, 0, 255).astype(np.uint8)

    return edgy_ims


def binarize_image(image, threshold):
    
    # Set values below switch_level to 0 and others to 255
    image[image < threshold] = 0
    image[image >= threshold] = 255
    image = np.clip(image, 0, 255).astype(np.uint8)

    '''
    # Invert the binary image
    inverted_image = cv2.bitwise_not(image)

    # Find contours of the inverted image (external contours)
    contours, _ = cv2.findContours(inverted_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Fill the external contours in the original binary image to close holes
    cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)
    '''
    return image

def fill_in_contours_1D(image):


    # Find contours of the inverted image (external contours)
    contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Draw contours on the binary image to close holes
    #print(" ---- ",image.shape, image.dtype, contours)

    image_copy = image.copy()
    cv2.drawContours(image_copy, contours, -1, 255, thickness=cv2.FILLED)
    #cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)

    # Merge the filled contours with the original binary image using bitwise OR
    #filled_image = cv2.bitwise_or(image, contour_filled_image)
    
    # Fill the external contours in the original binary image to close holes
    #cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)

    return image_copy


def fill_in_contours_3D(image):

    for channel in range(3):
        image[:,:,channel] = fill_in_contours_1D(image[:,:,channel])

    return image
    

def area_filter(image, threshold):

    # Find contours of the connected components
    contours, _ = cv2.findContours(image[:,:,1], cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    
    # Create a blank image to draw the filtered contours
    filtered_image = np.zeros_like(image[:,:,1])

    big_components = [contour for contour in contours 
                      if cv2.contourArea(contour) >= threshold]

    cv2.drawContours(filtered_image, big_components, -1, (255), -1)
    
    filtered_image_3D = np.stack((filtered_image, filtered_image, filtered_image), axis=-1)

    return filtered_image_3D


'''
def count_components(image):

    # Find contours of the connected components
    contours, _ = cv2.findContours(image[:,:,1], cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    return len(contours)
'''


def make_mask(src_image_np):
    temp_array_np = np.zeros_like(src_image_np)
    num_rows, num_cols = src_image_np.shape
    grid_size = 50
    grid_count = 0
    for curr_row in range(0,num_rows,grid_size):
        for curr_col in range(0,num_cols,grid_size):
            start_col, end_col, start_row, end_row = get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col)
            #print(np.mean(edges8[start_row:end_row,start_col:end_col]))
            if np.mean(src_image_np[start_row:end_row,start_col:end_col]) > 0:
                temp_array_np[start_row:end_row,start_col:end_col] = 255
                grid_count += 1

    #mask_array = np.where(temp_array_np==255, src_image_np,0)
    return temp_array_np, grid_count





def preprocess_image(image, data_dict, data_key, out_list,
                     func, **func_kw_args):

    if data_key in data_dict:
        print(f"Warning {data_key} has already been used as a data key")

    start_time = time()
    new_im = func(np.copy(image), **func_kw_args)

    #if func.__name__ != 'np_filter_function':
    #    name = func.__name__
    #elif 'np_func' in func_kw_args:
    #    name = 'np_' + func_kw_args['np_func'].__name__
    #else:
    #    print("Preprocess function not specified")
    name = data_key
        
    stat_str = np_stats_str(new_im)
    
    channel_list = [new_im[:,:,0], new_im[:,:,1], new_im[:,:,2]]
    title_list = ["_".join([name,out_list[0]]), "_".join([name,out_list[1]]), "_".join([name,out_list[2]])]
    num_keys = len(data_dict)
    data_dict[data_key] = {'channels':channel_list, 'titles':title_list, 'proc_num':num_keys}
    
    end_time = time()
    #print(name, new_im.dtype, stat_str, f"{end_time-start_time:.4f} sec")
    return new_im, data_dict


def show_processed_images(data_dict, grid_size=None):
    # Create a nx3 subplot layout
    n = len(data_dict)
    height = int(np.ceil(n *5))
    fig, axes = plt.subplots(n, 3, figsize=(15, height))

    # Force axes to be 2D even if n = 1
    axes = np.atleast_2d(axes)

    sorted_procs = sorted(data_dict, key=lambda k: data_dict[k]['proc_num'])

    for ctr, curr_proc in enumerate(sorted_procs):
        channels = data_dict[curr_proc]['channels']
        titles = data_dict[curr_proc]['titles']
        for i, (channel, title) in enumerate(zip(channels, titles)):
            axes[ctr, i].imshow(channel, cmap='gray', vmin=0, vmax=255)
            axes[ctr, i].set_title(title)
            if grid_size:
                axes[ctr, i].set_xticks(np.arange(0, channel.shape[1], grid_size))
                axes[ctr, i].set_yticks(np.arange(0, channel.shape[0], grid_size))
                axes[ctr, i].grid(color='cyan', linestyle='-', linewidth=1)
                axes[ctr, i].axis('on')

    # Adjust layout to prevent overlap
    plt.tight_layout()    


def make_frame_dict(src_dir_str):
    '''Return dict mapping each movie to its ordered frames'''
    src_dir = Path(src_dir_str)

    frame_dict = defaultdict(list)
    frame_list = sorted([x.name for x in src_dir.iterdir() 
                         if 'ipynb' not in x.name and x.is_file()])
    
    # Group images by prefix label
    for curr_img in frame_list:
        prefix = "_".join(curr_img.split("_")[0:9])
        frame_dict[prefix].append(curr_img)
    
    # Sort images for each prefix number by what I believe to be frame number
    for curr_movie in sorted(frame_dict.keys()):
        try:
            frame_dict[curr_movie] = sorted(frame_dict[curr_movie], 
                                            key=lambda x:int(x.split("_")[-1][:-4]))
        except:
            print(curr_movie,"---")
    return frame_dict

def find_fry(curr_image_path, predictor):
    out_list = ['raw','bckgnd','dframe']
    data_dict = dict()
    pred_boxes = get_gnd_truth_bboxes(curr_image_path, predictor, conf_thresh=0.5)
    curr_image_np = cv2.imread(curr_image_path)
    curr_image_np = cv2.cvtColor(curr_image_np, cv2.COLOR_BGR2RGB) 
    
    # Dummy Op - Return deep copy of original image 
    # so can be displayed & compared to preprocess results
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'orig', out_list,
                                                orig_image)
    
    # Remove large fish spotted by classifier
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'remove_fish', out_list,
                                                remove_big_fish, 
                                                pred_boxes=pred_boxes, 
                                                grid_size=50)
    
    # Remove anomalous dark patches
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'remove_dark', out_list,
                                                remove_dark_noise, 
                                                threshold_perc_list = [35], 
                                                frame_list = [1])
    
    # Smooth entire image with vectorized guided filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'guided_filter1', out_list,
                                                vectorized_guided_filter_rgb_independent, 
                                                radius = 8, eps = 0.01)
    
    # If top 10 pixels have greater than threshold intensity, enhance contrast with ImageEnhace
    # otherwise smooth with vectorized_guided_filter
    np_func_10 = nth_largest_element_wrapper(10) 
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'bright_reg_enhance', out_list,
                                                np_filter_function, np_func = np_func_10,#np.max,
                                                grid_size=50, thresholds=160,
                                                enhance_factor=2.0, radius=25, eps=10)
    
    # Enhance contrast with Clahe filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'clahe1', out_list,
                                                clahe_filter, 
                                                clip_limit=0.005)
    
    # Smooth entire image with vectorized guided filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'guided_filter2', out_list,
                                                vectorized_guided_filter_rgb_independent, 
                                                radius = 8, eps = 0.01)

    # Binarize image
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'binarize_image', out_list,
                                                binarize_image, 
                                                threshold=220)

    # Fill in contours
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'fill_in_contours', out_list,
                                                fill_in_contours_3D)

    # Focus on analyzing edges in images
    out_list = ['bckgnd','mag_grad', 'canny']
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'edgy_images', out_list,
                                                get_edgy_ims, 
                                                canny_lower=400, canny_upper=800)
     
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'edgy_images2', out_list,
                                                get_edgy_ims, 
                                                canny_lower=400, canny_upper=800)

    # Smooth entire image with vectorized guided filter    
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'fill_in_contours2', out_list,
                                                fill_in_contours_3D)

    # Smooth entire image with vectorized guided filter    
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'binarize_image2', out_list,
                                                binarize_image, 
                                                threshold=220)


    return curr_image_np, data_dict

def preprocess_fry_image(curr_image_np, predictor):

    out_list = ['raw','bckgnd','dframe']
    data_dict = dict()
    pred_boxes = get_gnd_truth_bboxes2(curr_image_np, predictor, conf_thresh=0.5)

    
    # Dummy Op - Return deep copy of original image 
    # so can be displayed & compared to preprocess results
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'orig', out_list,
                                                orig_image)
    
    # Remove large fish spotted by classifier
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'remove_fish', out_list,
                                                remove_big_fish, 
                                                pred_boxes=pred_boxes, 
                                                grid_size=50)
    
    # Remove anomalous dark patches
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'remove_dark', out_list,
                                                remove_dark_noise, 
                                                threshold_perc_list = [35], 
                                                frame_list = [1])
    
    # Smooth entire image with vectorized guided filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'guided_filter1', out_list,
                                                vectorized_guided_filter_rgb_independent, 
                                                radius = 8, eps = 0.01)
    
    # If top 10 pixels have greater than threshold intensity, enhance contrast with ImageEnhace
    # otherwise smooth with vectorized_guided_filter
    np_func_10 = nth_largest_element_wrapper(10) 
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'bright_reg_enhance', out_list,
                                                np_filter_function, np_func = np_func_10,#np.max,
                                                grid_size=50, thresholds=160,
                                                enhance_factor=2.0, radius=25, eps=10)
    
    # Enhance contrast with Clahe filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'clahe1', out_list,
                                                clahe_filter, 
                                                clip_limit=0.005)
    
    # Smooth entire image with vectorized guided filter
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'guided_filter2', out_list,
                                                vectorized_guided_filter_rgb_independent, 
                                                radius = 8, eps = 0.01)

    # Binarize image
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'binarize_image', out_list,
                                                binarize_image, 
                                                threshold=220)

    # Fill in contours
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'fill_in_contours', out_list,
                                                fill_in_contours_3D)

    # Focus on analyzing edges in images
    out_list = ['bckgnd','mag_grad', 'canny']
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'edgy_images', out_list,
                                                get_edgy_ims, 
                                                canny_lower=400, canny_upper=800)
     
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'edgy_images2', out_list,
                                                get_edgy_ims, 
                                                canny_lower=400, canny_upper=800)

    # Smooth entire image with vectorized guided filter    
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'fill_in_contours2', out_list,
                                                fill_in_contours_3D)

    # Smooth entire image with vectorized guided filter    
    curr_image_np, data_dict = preprocess_image(curr_image_np, data_dict, 'binarize_image2', out_list,
                                                binarize_image, 
                                                threshold=220)


    return curr_image_np, data_dict
    
#def get_components(image):
def get_contours(image):

    # Find contours of the connected components
    contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    return contours


def get_bounding_box_areas(contours):

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)
        
        # Draw the bounding box on the image (optional)
        #cv2.rectangle(image_with_boxes, (x, y), (x + w, y + h), (0, 255, 0), 2)  # Green box, thickness 2
        
        # Crop the region of interest (ROI) from the original grayscale image
        #roi = image[y:y+h, x:x+w]
        bbox_list.append(x*y)
    bbox_list = sorted(bbox_list)
    return bbox_list


def get_bounding_box_aspect_ratios(contours):

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)
        
        # Draw the bounding box on the image (optional)
        #cv2.rectangle(image_with_boxes, (x, y), (x + w, y + h), (0, 255, 0), 2)  # Green box, thickness 2
        
        # Crop the region of interest (ROI) from the original grayscale image
        #roi = image[y:y+h, x:x+w]
        bbox_list.append(x/(y+1e-6))
    bbox_list = sorted(bbox_list)
    return bbox_list

def contour_filter_area(contours, min_area):
    filtered_contours = []
    for curr in contours:
        x, y, w, h = cv2.boundingRect(curr)
        if x*y >= min_area:
            filtered_contours.append(curr)
    return filtered_contours

def contour_filter_aspect_ratio(contours, min_aspect_ratio):
    filtered_contours = []
    for curr in contours:
        x, y, w, h = cv2.boundingRect(curr)
        if x/(y+1e-6) >= min_aspect_ratio:
            filtered_contours.append(curr)
    return filtered_contours

# Example function to add text with consistent relative size and position
def rel_annotate_frame(image, text, relative_position=(.05,0.95), relative_font_size=0.001):
    height, width = image.shape[:2]  # Get the image dimensions
    
    # Calculate the absolute font size and position relative to the image dimensions
    font_scale = relative_font_size * min(width, height)  # Scale font size based on image size
    x_pos = int(relative_position[0] * width)   # X position as a fraction of image width
    y_pos = int(relative_position[1] * height)  # Y position as a fraction of image height
    #print(relative_position, height, width, x_pos, y_pos, font_scale)
    # Define other parameters for cv2.putText
    font = cv2.FONT_HERSHEY_SIMPLEX
    color = (255, 200, 200)  # White color (change as needed)
    thickness = 1
    
    # Add the text to the image
    cv2.putText(image, text, (x_pos, y_pos), font, font_scale, color, thickness, cv2.LINE_AA)

    return image


def make_fry_plot(fry_count_list, src_dir, out_dir, 
                  dataset, curr_movie_name, 
                  chosen_filter, chosen_channel, plot_title):

    curr_out_dir = out_dir / Path("Movies") / Path(dataset) / Path(curr_movie_name)
    curr_out_file = Path("_".join([chosen_filter, str(chosen_channel) + '.png']))
    fig, ax = plt.subplots()
    x = [y for y in range(len(fry_count_list))]
    # Plot the data
    ax.plot(x, fry_count_list, label='Fry Count')
    
    # Add labels and title
    ax.set_xlabel('Frame #')
    ax.set_ylabel('Fry')
    ax.set_title('Fry Count: ' + plot_title)
    
    # Add a legend
    ax.legend()
    
    # Display the plot
    print(f"Saving {curr_out_file}")
    fig.savefig(curr_out_dir / curr_out_file)


def make_bbox_hist_plot(bbox_list, src_dir, out_dir, 
                        dataset, curr_movie_name, file_suffix,
                        chosen_filter, chosen_channel, 
                        plot_title, xlabel,
                        bins=50 ):

    curr_out_dir = out_dir / Path("Movies") / Path(dataset) / Path(curr_movie_name)
    curr_out_file = Path("_".join([chosen_filter, str(chosen_channel) + file_suffix]))
    fig, ax = plt.subplots()

    # Plot the data
    # Create the histogram on the axis 'ax'
    n, bins, patches = ax.hist(bbox_list, bins=50, edgecolor='black')
    print(f"bins: {bins}")
    
    # Add labels and title
    ax.set_xlabel(xlabel)
    ax.set_ylabel('Count')
    ax.set_title('Fry Bbox Histogram: ' + plot_title)
    
    # Add a legend
    ax.legend()
    
    # Display the plot
    print(f"Saving {curr_out_file}")
    fig.savefig(curr_out_dir / curr_out_file)

'''
def make_movie_comparisons (frame_dict, src_dir, out_dir, dataset,
                            chosen_filter, channel, predictor=None, fps=10):
    ###Return dict mapping each movie to its ordered prediction frames
    annotation_frame_dict = defaultdict(list)
    num_movies = len(frame_dict)
    bad_file_list = []


    for ctr, curr_movie_name in enumerate(frame_dict):
        frame_list_orig = []
        frame_list_filt = []
        frame_list_overlay = []

        curr_out_file = Path(chosen_filter + '_overlay_' + curr_movie_name + '.mp4')
        curr_out_dir = out_dir / Path("Movies") / Path(dataset) / Path(curr_movie_name)
        curr_out_dir.mkdir(parents=True, exist_ok=True)
        print(curr_out_dir, (ctr+1), " of ", num_movies)
        for curr_img_name in tqdm(frame_dict[curr_movie_name]):

            curr_img_path = os.path.join(src_dir, curr_img_name)
            curr_img = cv2.imread(curr_img_path)

            curr_img_shape = curr_img.shape
            text_row_center = int(0.90 * curr_img_shape[0])
            text_col_center = int(0.125 * curr_img_shape[1])

            
            fry_image, fry_data_dict = find_fry(curr_img_path, predictor)
            chosen_image = fry_data_dict[chosen_filter]['channels'][channel]
            num_fry = count_components(chosen_image)
            
            final_image_orig = np.stack((curr_img[:,:,1], curr_img[:,:,1], curr_img[:,:,1]), axis=-1)
            final_image_orig = rel_annotate_frame(final_image_orig, "Bckgnd Avg.", (.05,0.95) )
            frame_list_orig.append(final_image_orig)

            final_image_filt = np.stack((chosen_image, chosen_image, chosen_image), axis=-1)
            final_image_filt = rel_annotate_frame(final_image_filt, "Highlit Fry", (.05,0.95) )
            frame_list_filt.append(final_image_filt)

            overlay_image = np.maximum(curr_img[:,:,1], chosen_image)
            final_image_overlay = np.stack((overlay_image, overlay_image, overlay_image), axis=-1)
            final_image_overlay = rel_annotate_frame(final_image_overlay, "Fry Count", (.05,0.95) )
            final_image_overlay = rel_annotate_frame(final_image_overlay, str(num_fry), (.05,0.975) )
            frame_list_overlay.append(final_image_overlay)

        # Create a video clip from the image frames
        clip_orig = ImageSequenceClip(frame_list_orig, fps=fps)  # Adjust fps as needed
        clip_filt = ImageSequenceClip(frame_list_filt, fps=fps)  # Adjust fps as needed
        clip_overlay = ImageSequenceClip(frame_list_overlay, fps=fps)  # Adjust fps as needed

        # Get the height of the clips (they should all have the same height)
        clip_height = clip_orig.h

        # Create a narrow vertical spacer (e.g., 10px wide) with the same height and black color
        spacer = ColorClip(size=(10, clip_height), color=(0, 0, 0), duration=clip_orig.duration)

        # Arrange the clips side by side with spacers in between
        final_clip = clips_array([[clip_orig, spacer, clip_filt, spacer, clip_overlay]])

        # Save the video clip
        print(f"Saving {curr_out_dir / curr_out_file}")
        out_file_posix = curr_out_dir / curr_out_file
        final_clip.write_videofile(out_file_posix.as_posix())

            
            # Save image            
            #outfile = curr_out_dir / Path("Pred_" + curr_img_name)
            #cv2.imwrite('/'.join(outfile.parts),final_image)
            #annotation_frame_dict[curr_movie_name].append('/'.join(outfile.parts))
    return #frame_list #annotation_frame_dict
'''

'''
def annotate_frame(final_image, text, col_center, row_center):
    # Text parameters
    text = text 
    fontFace = cv2.FONT_HERSHEY_SIMPLEX
    fontScale = 0.65
    color = (255,200,200)  
    thickness = 1
    
    # Get image dimensions
    _, width, _ = final_image.shape
    
    # Calculate center coordinates for top-center position
    text_center_col = col_center #int(width / 2)
    text_center_row = row_center  # Adjust y-coordinate as needed (distance from top edge)
    
    # Get text size using getTextSize
    (text_width, text_height), _ = cv2.getTextSize(text, fontFace, fontScale, thickness)
    
    # Adjust center_y based on text height for better alignment
    text_center_row = text_center_row + text_height // 2  # Add half of text height
    
    # Bottom-left corner coordinates relative to center
    text_org = (text_center_col - text_width // 2, text_center_row)  
    print(f"final image shape = {final_image.shape}")
    print(f"text_org = {text_org}")
    
    # Write text on image
    #frame=frame.copy()
    cv2.putText(final_image, text, text_org, 
                fontFace, fontScale, 
                color, thickness, cv2.LINE_AA)
    return final_image
'''